#include "encfs.h"

// Qt
#include <QDir>
#include <QProcess>
#include <QRegularExpression>
#include <QTextStream>
#include <QTemporaryDir>

#include "stash.h"

Encfs::Encfs(QObject *parent) :
    QObject(parent),
    m_encfsProcess(),
    m_userCanceled(false)
{
}

Encfs::~Encfs()
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        m_encfsProcess->terminate();
        return;
    }
}

bool Encfs::isUserAllowOtherInFuseConf()
{
    QFile file(QStringLiteral("/etc/fuse.conf"));
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        return false;
    }

    QRegularExpression re(QStringLiteral("^\\s*user_allow_other"));
    QTextStream in(&file);
    while (!in.atEnd()) {
        QString line = in.readLine();
        if (line.contains(re)) {
            return true;
        }
    }

    return false;
}

void Encfs::mount(const QSharedPointer<const Stash> stash, const QString &password)
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        emit mountFinished(false, tr("The stash is busy"));
        return;
    }

    if (stash->allowOtherUsers() || stash->allowRootAccess()) {
        if (!isUserAllowOtherInFuseConf()) {
            emit mountFinished(false, tr("Please enable the option \"user_allow_other\" in the file \"/etc/fuse.conf\" and make sure that the file is readable by you."));
            return;
        }
    }

    m_encfsProcess = QSharedPointer<QProcess>::create();
    QStringList args;

    args << QLatin1String("-S");

    if (stash->unmountIdle() && (stash->unmountIdleTime() > 0)) {
        args << QString(QLatin1String("--idle=%1")).arg(QString::number(stash->unmountIdleTime()));
    }

    if (!stash->additionalMountOptions().isEmpty()) {
        args << parseCombinedArgString(stash->additionalMountOptions());
    }

    // Create directory if it doesn't exist and set permissions to 700
    QDir mountDir(stash->mountDir());
    if (!mountDir.exists()) {
       if (!mountDir.mkpath(QLatin1String("."))) {
           emit mountFinished(false, tr("The mount directory couldn't be created."));
           return;
       }
    } else {
        if (stash->allowNonEmptyMountDirectory()) {
            args << QLatin1String("-o") << QLatin1String("nonempty");
        } else {
            if (!mountDir.isEmpty()) {
                emit mountFinished(false, tr("The mount directory isn't empty.\n"
                                             "Did you accidentally save unencrypted files there?\n"
                                             "You can explicitely allow it in the stash configuration\n"
                                             "if you know what you're doing though."));
                return;
            }
        }
    }
    QFile::setPermissions(stash->mountDir(), QFileDevice::ReadOwner | QFileDevice::WriteOwner | QFileDevice::ExeOwner);

    args << stash->stashDir() << stash->mountDir();

    if (stash->allowOtherUsers() || stash->allowRootAccess()) {
        args << QLatin1String("--");
        if (stash->allowOtherUsers()) {
            args << QLatin1String("-o") << QLatin1String("allow_other");
        }
        if (stash->allowRootAccess()) {
            args << QLatin1String("-o") << QLatin1String("allow_root");
        }
    }


    if (!stash->encfsConfigFile().isEmpty()) {
        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        env.insert(QLatin1String("ENCFS6_CONFIG"), stash->encfsConfigFile());
        m_encfsProcess->setProcessEnvironment(env);
    }

    m_encfsProcess->setProgram(QLatin1String("encfs"));
    m_encfsProcess->setArguments(args);

    connect(m_encfsProcess.data(), QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            [this, password](int exitCode, QProcess::ExitStatus exitStatus) {

        if (m_userCanceled) {
            m_userCanceled = false;
            emit mountFinished(false, tr("Mounting was cancelled."));
        } else {
            if ((exitStatus == QProcess::NormalExit) && (exitCode == 0)) {
                emit mountFinished(true, password);
            } else {
                emit mountFinished(false, m_encfsProcess->canReadLine() ? m_encfsProcess->readAll().trimmed() : tr("An unknown error occurred"));
            }
        }
    });

    QTextStream out(stdout);
    out << QLatin1String("Command: \"") + m_encfsProcess->program() + QLatin1String("\" \"") + m_encfsProcess->arguments().join("\" \"") + QLatin1String("\"") << endl;
    m_encfsProcess->start();

    if (!m_encfsProcess->waitForStarted()) {
        emit mountFinished(false, tr("Failed to execute encfs.\n"
                                     "Please make sure that it is installed and in your PATH."));
        return;
    }
    m_encfsProcess->write(password.toUtf8()+"\n");
    m_encfsProcess->closeWriteChannel();
}

void Encfs::unmount(const QSharedPointer<const Stash> stash)
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        emit unmountFinished(false, tr("The stash is busy"));
        return;
    }
    m_encfsProcess = QSharedPointer<QProcess>::create();
    QStringList args;

    args << QLatin1String("-u");
    args << stash->mountDir();

    m_encfsProcess->setProgram(QLatin1String("fusermount"));
    m_encfsProcess->setArguments(args);

    connect(m_encfsProcess.data(), QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            [this](int exitCode, QProcess::ExitStatus exitStatus) {

        if (m_userCanceled) {
            m_userCanceled = false;
            emit unmountFinished(false, tr("Unmounting was cancelled."));
        } else {
            if ((exitStatus == QProcess::NormalExit) && (exitCode == 0)) {
                emit unmountFinished(true, QStringLiteral(""));
            } else {
                emit unmountFinished(false, m_encfsProcess->canReadLine() ? m_encfsProcess->readAll().trimmed() : tr("An unknown error occurred.\n"
                                                                                                                     "Is the stash already unmounted?"));
            }
        }
    });

    QTextStream out(stdout);
    out << QLatin1String("Command: \"") + m_encfsProcess->program() + QLatin1String("\" \"") + m_encfsProcess->arguments().join("\" \"") + QLatin1String("\"") << endl;
    m_encfsProcess->start();

    if (!m_encfsProcess->waitForStarted()) {
        emit unmountFinished(false, tr("Failed to execute fusermount.\n"
                                       "Please make sure that it is installed and in your PATH."));
        return;
    }
}

void Encfs::create(const QSharedPointer<const Stash> stash, const QString &password, CreateStashFlags flags)
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        emit createFinished(false, tr("The stash is busy"));
        return;
    }

    if (stash->allowOtherUsers() || stash->allowRootAccess()) {
        if (!isUserAllowOtherInFuseConf()) {
            emit mountFinished(false, tr("Please enable the option \"user_allow_other\" in the file \"/etc/fuse.conf\" and make sure that the file is readable by you."));
            return;
        }
    }

    m_encfsProcess = QSharedPointer<QProcess>::create();
    QStringList args;
    args << QLatin1String("-S");

    // Create directory if it doesn't exist and set permissions to 700
    QDir mountDir(stash->mountDir());
    if (!mountDir.exists()) {
       if (!mountDir.mkpath(QLatin1String("."))) {
           emit createFinished(false, tr("The mount directory couldn't be created."));
           return;
       }
    } else {
        if (stash->allowNonEmptyMountDirectory()) {
            args << QLatin1String("-o") << QLatin1String("nonempty");
        } else {
            if (!mountDir.isEmpty()) {
                emit createFinished(false, tr("The mount directory isn't empty.\n"
                                              "Did you accidentally save unencrypted files there?\n"
                                              "You can explicitely allow it in the stash configuration\n"
                                              "if you know what you're doing though."));
                return;
            }
        }
    }
    QFile::setPermissions(stash->mountDir(), QFileDevice::ReadOwner | QFileDevice::WriteOwner | QFileDevice::ExeOwner);

    QDir stashDir(stash->stashDir());
    if (!stashDir.exists()) {
       if (!stashDir.mkpath(QLatin1String("."))) {
           emit createFinished(false, tr("The stash directory couldn't be created."));
           return;
       }
    }/* else {
        if (!stashDir.isEmpty()) {
            emit createFinished(false, tr("The stash directory isn't empty.\n"
            return;
        }
    }*/
    QFile::setPermissions(stash->stashDir(), QFileDevice::ReadOwner | QFileDevice::WriteOwner | QFileDevice::ExeOwner);

    QTemporaryDir *tempDir = nullptr;
    QString encfsConfigFile = stash->encfsConfigFile();
    if (!encfsConfigFile.isEmpty()) {
        /* create stash in a temporary directory, then move the
           encfs config file to the desired location when finished */
        tempDir = new QTemporaryDir;
        if (!tempDir->isValid()) {
            emit createFinished(false, tr("Failed to create a temporary directory for the custom encfs config file."));
            return;
        } else {
            args << tempDir->path();
        }

        //QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        //env.insert(QLatin1String("ENCFS6_CONFIG"), stash->encfsConfigFile());
        //m_encfsProcess->setProcessEnvironment(env);
    } else {
        args << stash->stashDir();
    }

    args << stash->mountDir();


    m_encfsProcess->setProgram(QLatin1String("encfs"));
    m_encfsProcess->setArguments(args);

    connect(m_encfsProcess.data(), QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            [this, password, tempDir, encfsConfigFile](int exitCode, QProcess::ExitStatus exitStatus) {
        if (m_userCanceled) {
            m_userCanceled = false;
            delete(tempDir);
            emit createFinished(false, tr("Stash creation was cancelled."));
        } else {
            if ((exitStatus == QProcess::NormalExit) && (exitCode == 0)) {
                if (tempDir != nullptr) {
                    QTextStream out(stdout);
                    out << "Moving encfs config file to " << encfsConfigFile << endl;
                    QString filename = tempDir->filePath(QStringLiteral(".encfs6.xml"));
                    QDir(QFileInfo(encfsConfigFile).dir()).mkpath(QLatin1String("."));
                    QFile(filename).rename(encfsConfigFile);
                }
                delete(tempDir);
                emit createFinished(true, password);
            } else {
                delete(tempDir);
                emit createFinished(false, m_encfsProcess->canReadLine() ? m_encfsProcess->readAll().trimmed() : QStringLiteral(""));
            }
        }
    });

    m_encfsProcess->setProcessChannelMode(QProcess::ForwardedChannels);

    QTextStream out(stdout);
    out << QLatin1String("Command: \"") + m_encfsProcess->program() + QLatin1String("\" \"") + m_encfsProcess->arguments().join("\" \"") + QLatin1String("\"") << endl;
    m_encfsProcess->start();

    if (!m_encfsProcess->waitForStarted()) {
        emit createFinished(false, tr("Failed to execute encfs.\n"
                                      "Please make sure that it is installed and in your PATH."));
        return;
    }

    if (flags == 0) {
        m_encfsProcess->write("p\n");      // Paranoia mode
        m_encfsProcess->write(password.toUtf8()+"\n");  // password
        m_encfsProcess->write(password.toUtf8()+"\n");  // verify password
        m_encfsProcess->closeWriteChannel();
    } else if (flags == Encfs::DisableFilenameEncryption) {
        m_encfsProcess->write("x\n");      // Expert mode
        m_encfsProcess->write("1\n");      // Cipher algorithm: 1. AES, 2. Blowfish
        m_encfsProcess->write("256\n");    // key size in bits
        m_encfsProcess->write("1024\n");   // block size in bytes
        m_encfsProcess->write("3\n");      /* filename encoding algorithm 1. Block : Block encoding, hides file name size somewhat
                                              2. Block32 : Block encoding with base32 output for case-sensitive systems
                                              3. Null : No encryption of filenames
                                              4. Stream : Stream encoding, keeps filenames as short as possible*/
        m_encfsProcess->write("y\n");      // filename initialization vector chaining
        m_encfsProcess->write("y\n");      // per-file initialization vectors
        m_encfsProcess->write("n\n");      // filename to IV header chaining?
        m_encfsProcess->write("n\n");      // block authentication code headers
        m_encfsProcess->write("8\n");      // Add random bytes to each block header
        m_encfsProcess->write("n\n");      // file-hole pass-through

        m_encfsProcess->write(password.toUtf8()+"\n");  // password
        m_encfsProcess->write(password.toUtf8()+"\n");  // verify password
        m_encfsProcess->closeWriteChannel();
    }
}

void Encfs::changePassword(const QSharedPointer<const Stash> stash, const QString &oldPassword, const QString &newPassword)
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        emit changePasswordFinished(false, tr("The stash is busy"));
        return;
    }

    m_encfsProcess = QSharedPointer<QProcess>::create();
    QStringList args;
    args << QLatin1String("passwd");
    args << stash->stashDir();

    if (!stash->encfsConfigFile().isEmpty()) {
        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        env.insert(QLatin1String("ENCFS6_CONFIG"), stash->encfsConfigFile());
        m_encfsProcess->setProcessEnvironment(env);
    }

    connect(m_encfsProcess.data(), QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            [this, newPassword](int exitCode, QProcess::ExitStatus exitStatus) {

        if (m_userCanceled) {
            m_userCanceled = false;
            emit changePasswordFinished(false, tr("The password change was cancelled."));
        } else {
            if ((exitStatus == QProcess::NormalExit) && (exitCode == 0)) {
                emit changePasswordFinished(true, newPassword);
            } else {
                // Only show last line as the message
                QString message = QLatin1String();
                while (m_encfsProcess->canReadLine()) {
                    message = m_encfsProcess->readLine();
                }
                message = message.trimmed();
                emit changePasswordFinished(false, message);
            }
        }
    });

    m_encfsProcess->setProgram(QLatin1String("encfsctl"));
    m_encfsProcess->setArguments(args);

    QTextStream out(stdout);
    out << QLatin1String("Command: \"") + m_encfsProcess->program() + QLatin1String("\" \"") + m_encfsProcess->arguments().join("\" \"") + QLatin1String("\"") << endl;
    m_encfsProcess->start();

    if (!m_encfsProcess->waitForStarted()) {
        emit changePasswordFinished(false, tr("Failed to execute encfsctl.\n"
                                     "Please make sure that it is installed and in your PATH."));
        return;
    }

    m_encfsProcess->write(oldPassword.toUtf8()+"\n");
    m_encfsProcess->write(newPassword.toUtf8()+"\n");
    m_encfsProcess->write(newPassword.toUtf8()+"\n");
    m_encfsProcess->closeWriteChannel();
}

QString Encfs::info(const QSharedPointer<const Stash> stash)
{
    QProcess encfsProcess;
    QStringList args;

    if (!stash->encfsConfigFile().isEmpty()) {
        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        env.insert(QLatin1String("ENCFS6_CONFIG"), stash->encfsConfigFile());
        encfsProcess.setProcessEnvironment(env);
    }

    args << QLatin1String("info");
    args << stash->stashDir();

    encfsProcess.setProgram(QLatin1String("encfsctl"));
    encfsProcess.setArguments(args);
    encfsProcess.setProcessChannelMode(QProcess::ForwardedErrorChannel);

    QTextStream out(stdout);
    out << QLatin1String("Command: \"") +encfsProcess.program() + QLatin1String("\" \"") + encfsProcess.arguments().join("\" \"") + QLatin1String("\"") << endl;
    encfsProcess.start();

    encfsProcess.waitForFinished();
    if (encfsProcess.exitStatus() == QProcess::NormalExit) {
        return QString(encfsProcess.readAllStandardOutput()).trimmed();
    } else {
        return tr("Error fetching stash information");
    }
}

void Encfs::cancel()
{
    if (!m_encfsProcess.isNull() && (m_encfsProcess->state() == QProcess::Running)) {
        m_userCanceled = true;
        m_encfsProcess->terminate();
    }
}

QStringList Encfs::parseCombinedArgString(const QString &arguments)
{
    QStringList args;
    QString tmp;
    int quoteCount = 0;
    bool inQuote = false;
    // handle quoting. tokens can be surrounded by double quotes
    // "hello world". three consecutive double quotes represent
    // the quote character itself.
    for (int i = 0; i < arguments.size(); ++i)
    {
        if (arguments.at(i) == QLatin1Char('"'))
        {
            ++quoteCount;
            if (quoteCount == 3)
            {
                // third consecutive quote
                quoteCount = 0;
                tmp += arguments.at(i);
            }
            continue;
        }
        if (quoteCount)
        {
            if (quoteCount == 1)
                inQuote = !inQuote;
            quoteCount = 0;
        }
        if (!inQuote && arguments.at(i).isSpace())
        {
            if (!tmp.isEmpty())
            {
                args += tmp;
                tmp.clear();
            }
        }
        else
        {
            tmp += arguments.at(i);
        }
    }
    if (!tmp.isEmpty())
        args += tmp;
    return args;
}
