#include "pager.h"
#include "ui_pager.h"

// Qt
#include <QStyledItemDelegate>
#include <QAbstractListModel>
#include <QDebug>
#include <QPainter>
#include <QWizardPage>

// Core
#include "Core/settings.h"
#include "Core/stash.h"
#include "Core/stashmanager.h"

class Pager::MenuListModel : public QAbstractListModel
{
    Q_OBJECT

public:
    explicit MenuListModel(QObject *parent = nullptr) :
        QAbstractListModel(parent),
        m_stackedWidget(qobject_cast<Pager*>(this->parent())->ui->stackedWidget)
    {
        connect(qobject_cast<Pager*>(this->parent()), &Pager::pageAdded, [this]() {
            beginResetModel();
            endResetModel();
        });
    }

    virtual int rowCount(const QModelIndex &parent) const override
    {
        Q_UNUSED(parent)
        return m_stackedWidget->count();
    }

    virtual QVariant data(const QModelIndex &index, int role) const override
    {
        switch (role) {
        case Qt::DisplayRole:
            return qobject_cast<QWizardPage*>(m_stackedWidget->widget(index.row()))->title();
            break;
        case Qt::DecorationRole:
            return qobject_cast<QWizardPage*>(m_stackedWidget->widget(index.row()))->pixmap(QWizard::LogoPixmap);
            break;
        }
        return QVariant();
    }

private:
    const QStackedWidget *m_stackedWidget;
};

class Pager::MenuItemDelegate : public QStyledItemDelegate
{
    Q_OBJECT

public:
    explicit MenuItemDelegate(QObject *parent = nullptr) : QStyledItemDelegate(parent)
    {

    }

    virtual void paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const override
    {
        //painter->fillRect(option.rect.adjusted(100, 0,0,0), QColor(Qt::white));
       // qDebug() << painter->
        //QPainter p(qobject_cast<Pager*>(parent()));
        //p.fillRect(option.rect.adjusted(100, 0,0,0), QColor(Qt::white));

        QStyleOptionViewItem opt(option);
        opt.rect.adjust(0,0,0,-5);
        opt.state &= ~QStyle::State_HasFocus;
        QStyledItemDelegate::paint(painter, opt, index);
    }

    virtual QSize sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const override
    {
        QSize itemSize = QStyledItemDelegate::sizeHint(option, index);
        itemSize.setHeight(itemSize.height()+10);
        m_minSize = m_minSize.expandedTo(itemSize);
        qobject_cast<Pager*>(this->parent())->ui->listView->setMinimumWidth(m_minSize.width()+6);
        qobject_cast<Pager*>(this->parent())->adjustSize();
        return m_minSize;
    }

public:
    mutable QSize m_minSize;
};

Pager::Pager(QSharedPointer<Stash> stash, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::Pager),
    m_stash(stash)
{
    ui->setupUi(this);
    if (!Settings::instance().showSidebar()) {
        ui->listView->setVisible(false);
    }

    setWindowTitle(stash->shortName());

    connect(&Settings::instance(), &Settings::showSidebarChanged, this, &Pager::sidebarToggled);

    ui->listView->setModel(new MenuListModel(this));
    ui->listView->setItemDelegateForColumn(0, new MenuItemDelegate(this));
    connect(ui->listView->selectionModel(), &QItemSelectionModel::selectionChanged, [this](const QItemSelection &selected, const QItemSelection &deselected) {
        Q_UNUSED(deselected)
        ui->stackedWidget->setCurrentIndex(selected.at(0).indexes().at(0).row());
        adjustSize();
    });

    QPalette pal = palette();
    pal.setColor(QPalette::Highlight, pal.color(QPalette::Window));
    pal.setColor(QPalette::HighlightedText, pal.color(QPalette::WindowText));
    ui->listView->setPalette(pal);
}

Pager::Pager(const QString &title, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::Pager)
{
    ui->setupUi(this);
    if (!Settings::instance().showSidebar()) {
        ui->listView->setVisible(false);
    }

    setWindowTitle(title);

    connect(&Settings::instance(), &Settings::showSidebarChanged, this, &Pager::sidebarToggled);

    ui->listView->setModel(new MenuListModel(this));
    ui->listView->setItemDelegateForColumn(0, new MenuItemDelegate(this));
    connect(ui->listView->selectionModel(), &QItemSelectionModel::selectionChanged, [this](const QItemSelection &selected, const QItemSelection &deselected) {
        Q_UNUSED(deselected)
        ui->stackedWidget->setCurrentIndex(selected.at(0).indexes().at(0).row());
        adjustSize();
    });

    QPalette pal = palette();
    pal.setColor(QPalette::Highlight, pal.color(QPalette::Window));
    pal.setColor(QPalette::HighlightedText, pal.color(QPalette::WindowText));
    ui->listView->setPalette(pal);
}

Pager::~Pager()
{
    delete ui;
}

void Pager::addPage(QWizardPage *wizardPage)
{
    int index = ui->stackedWidget->addWidget(wizardPage);
    if (index == 0) {
        m_firstWidget = wizardPage;
    }
    emit pageAdded();
}

void Pager::insertPage(int index, QWizardPage *wizardPage)
{
    ui->stackedWidget->insertWidget(index, wizardPage);
    emit pageAdded();
}

void Pager::selectPage(int pageNr)
{
    ui->listView->selectionModel()->select(ui->listView->model()->index(pageNr,0), QItemSelectionModel::SelectCurrent);
}

void Pager::adjustSize()
{
    int minHeight = 1;
    if (Settings::instance().showSidebar()) {
        minHeight = qobject_cast<MenuItemDelegate*>(ui->listView->itemDelegateForColumn(0))->m_minSize.height() * ui->stackedWidget->count();
        minHeight = qMax(minHeight, 400);
    }

    setMaximumSize(99999,99999);
    setMinimumSize(0, minHeight >= 1 ? minHeight : 1);
    QDialog::adjustSize();
    setMaximumSize(size());
    setMinimumSize(size());
}

void Pager::sidebarToggled(bool showSidebar)
{
    ui->stackedWidget->setCurrentWidget(m_firstWidget);
    int oldWith = size().width();
    int oldHeight = size().height();
    ui->listView->setVisible(showSidebar);
    adjustSize();
    move(pos().x() - (width()-oldWith)/2, pos().y() - (height()-oldHeight)/2);
}

void Pager::reject()
{
    if (!m_stash.isNull() && m_stash->isBusy()) {
        StashManager::instance().cancel(m_stash);
    } else {
        QDialog::reject();
    }
}

#include "pager.moc"
