#include "configurestashexpertpage.h"
#include "ui_configurestashexpertpage.h"

// Qt
#include <QFileDialog>
#include <QIcon>

// Core
#include "Core/encfs.h"
#include "Core/helpers.h"
#include "Core/settings.h"
#include "Core/stash.h"

#include <MHelpBrowser>

ConfigureStashExpertPage::ConfigureStashExpertPage(const QSharedPointer<Stash> &stash, QWidget *parent) :
    QWizardPage(parent),
    ui(new Ui::ConfigureStashExpertPage),
    m_stash(stash),
    m_initialized(false),
    m_size(QSize(1,1))
{
    ui->setupUi(this);

    setPixmap(QWizard::LogoPixmap, QIcon::fromTheme(QLatin1String("document-edit")).pixmap(24,24));

    ui->label_help->setText(QString(QLatin1String("<html><a href=\"stashadvancedconfiguration.html\">%1</a></html>")).arg(tr("Help")));

    ui->lineEdit_encfsConfigFile->setMinimumHeight(ui->button_encfsConfigFile->height()-6);
    ui->lineEdit_keyringName->setMinimumHeight(ui->button_encfsConfigFile->height()-6);
}

ConfigureStashExpertPage::~ConfigureStashExpertPage()
{
    delete ui;
}

void ConfigureStashExpertPage::setVisible(bool visible)
{
    if (visible && !m_initialized) {
        m_initialized = true;

        ui->button_sidebar->setChecked(Settings::instance().showSidebar());
        connect(ui->button_sidebar, &QPushButton::toggled, &Settings::instance(), &Settings::setShowSidebar);
        connect(&Settings::instance(), &Settings::showSidebarChanged, ui->button_sidebar, &QPushButton::setChecked);

        ui->lineEdit_encfsConfigFile->setText(m_stash->encfsConfigFile());
        connect(m_stash.data(), &Stash::encfsConfigFileChanged, this, &ConfigureStashExpertPage::encfsConfigFileChanged);
        connect(ui->lineEdit_encfsConfigFile, &QLineEdit::textChanged, m_stash.data(), &Stash::setEncfsConfigFile);

        ui->lineEdit_keyringName->setText(m_stash->keyringName());
        connect(m_stash.data(), &Stash::keyringNameChanged, this, &ConfigureStashExpertPage::keyringNameChanged);
        connect(ui->lineEdit_keyringName, &QLineEdit::textChanged, m_stash.data(), &Stash::setKeyringName);

        ui->checkBox_allowNonEmptyMountDirectory->setChecked(m_stash->allowNonEmptyMountDirectory());
        connect(m_stash.data(), &Stash::allowNonEmptyMountDirectoryChanged, ui->checkBox_allowNonEmptyMountDirectory, &QCheckBox::setChecked);
        connect(ui->checkBox_allowNonEmptyMountDirectory, &QCheckBox::toggled, m_stash.data(), &Stash::setAllowNonEmptyMountDirectory);

        bool otherUsersAvailable = Encfs::isUserAllowOtherInFuseConf();
        ui->checkBox_allowOtherUsers->setChecked(m_stash->allowOtherUsers());
        ui->checkBox_allowOtherUsers->setEnabled(otherUsersAvailable);
        connect(m_stash.data(), &Stash::allowOtherUsersChanged, ui->checkBox_allowOtherUsers, &QCheckBox::setChecked);
        connect(ui->checkBox_allowOtherUsers, &QCheckBox::toggled, m_stash.data(), &Stash::setAllowOtherUsers);

        ui->checkBox_allowRootAccess->setChecked(m_stash->allowRootAccess());
        ui->checkBox_allowRootAccess->setEnabled(otherUsersAvailable);
        connect(m_stash.data(), &Stash::allowRootAccessChanged, ui->checkBox_allowRootAccess, &QCheckBox::setChecked);
        connect(ui->checkBox_allowRootAccess, &QCheckBox::toggled, m_stash.data(), &Stash::setAllowRootAccess);

        if (m_stash->isInitialized()) {
            ui->line_createOptions->setVisible(false);
            ui->groupBox_createOptions->setVisible(false);
        } else {
            ui->checkBox_encryptFilenames->setChecked(!m_stash->m_createFlags.testFlag(Encfs::DisableFilenameEncryption));
            connect(ui->checkBox_encryptFilenames, &QCheckBox::toggled, [this](bool checked) {
                m_stash->m_createFlags.setFlag(Encfs::DisableFilenameEncryption, !checked);
            });
        }
    }

    if (visible) {
        m_size = QWizardPage::minimumSizeHint();
    } else {
        m_size = QSize(1,1);
    }

    QWizardPage::setVisible(visible);
}

QSize ConfigureStashExpertPage::sizeHint() const
{
    return m_size;
}

QSize ConfigureStashExpertPage::minimumSizeHint() const
{
    return m_size;
}

void ConfigureStashExpertPage::on_button_encfsConfigFile_clicked()
{
    QFileDialog fileDialog(this);
    fileDialog.setDirectory(ui->lineEdit_encfsConfigFile->text().isEmpty() ? Helpers::pathToAbsolute(Settings::instance().baseDir()) : ui->lineEdit_encfsConfigFile->text());
    fileDialog.setFileMode(QFileDialog::AnyFile);
    fileDialog.setOptions(QFileDialog::DontConfirmOverwrite);
    if (m_stash->isInitialized()) {
        fileDialog.setAcceptMode(QFileDialog::QFileDialog::AcceptOpen);
    } else {
        fileDialog.setAcceptMode(QFileDialog::QFileDialog::AcceptSave);
    }

    if(fileDialog.exec()) {
       ui->lineEdit_encfsConfigFile->setText(fileDialog.selectedFiles().first());
    }
}

void ConfigureStashExpertPage::encfsConfigFileChanged(const QString &encfsConfigFile)
{
    if (ui->lineEdit_encfsConfigFile->text() == encfsConfigFile) {
        return;
    }

    ui->lineEdit_encfsConfigFile->setText(encfsConfigFile);
}

void ConfigureStashExpertPage::keyringNameChanged(const QString &keyringName)
{
    if (ui->lineEdit_keyringName->text() == keyringName) {
        return;
    }

    ui->lineEdit_keyringName->setText(keyringName);
}

void ConfigureStashExpertPage::on_label_help_linkActivated(const QString &link)
{
    MHelpBrowser::open(link);
}
