#include "stashgroupwidget.h"
#include "ui_stashgroupwidget.h"

// Qt
#include <QAbstractListModel>
#include <QDebug>
#include <QEvent>
#include <QKeyEvent>
#include <QMetaObject>
#include <QMimeData>
#include <QPainter>
#include <QProxyStyle>
#include <QStringListModel>

// Core
#include "Core/settings.h"
#include "Core/stash.h"
#include "Core/stashmanager.h"

#include "stashmimedata.h"


/*!
 * \brief The MyProxyStyle class
 *
 * Shades dropped onto items in a light gray and the move indicator (the line
 * inbetween the items) also in a light gray
 */
class MyProxyStyle: public QProxyStyle{
public:
    MyProxyStyle(QStyle* style = 0) : QProxyStyle(style) {}

    void drawPrimitive ( PrimitiveElement element, const QStyleOption * option, QPainter * painter, const QWidget * widget = 0 ) const {
        qDebug() << option->state;
        if (element == QStyle::PE_IndicatorItemViewItemDrop && !option->rect.isNull()){
            if (option->rect.height() == 0) {
                painter->fillRect(option->rect.adjusted(0,0,0,1), QColor(0,0,0,100));
            } else {
                QColor col =  widget->palette().highlight().color();
                col.setAlphaF(0.15);
                painter->fillRect(option->rect, col);
            }
            return;
        }
        QProxyStyle::drawPrimitive(element, option, painter, widget);
    }
};


class HeaderNavigationFilter : public QObject
{
    Q_OBJECT
public:
    explicit HeaderNavigationFilter(QObject *parent = nullptr) : QObject(parent) {}

    bool eventFilter(QObject *obj, QEvent *event) {
        if (event->type() == QEvent::KeyPress) {
            StashgroupWidget *wid = qobject_cast<StashgroupWidget*>(parent());
            QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
            QTableView *view = qobject_cast<QTableView*>(obj);

            if (keyEvent->key() == Qt::Key_Down) {
                if (view->currentIndex().row() == 0) {
                    wid->ui->defaultTableView->selectRow(0);
                    wid->ui->defaultTableView->setFocus();
                    return true;
                }
            } else if (keyEvent->key() == Qt::Key_Tab) {
                view->selectionModel()->clearSelection();
                wid->ui->button_add->setFocus();
                return true;
            }
        }
        return QObject::eventFilter(obj, event);
    }
};


class DefaultNavigationFilter : public QObject
{
    Q_OBJECT
public:
    explicit DefaultNavigationFilter(QObject *parent = nullptr) : QObject(parent) {}

    bool eventFilter(QObject *obj, QEvent *event) {
        if (event->type() == QEvent::KeyPress) {
            StashgroupWidget *wid = qobject_cast<StashgroupWidget*>(parent());
            QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
            QTableView *view = qobject_cast<QTableView*>(obj);

            if (keyEvent->key() == Qt::Key_Up) {
                if (view->currentIndex().row() == 0) {
                    wid->ui->headerTableView->selectRow(0);
                    wid->ui->headerTableView->setFocus();
                    return true;
                }
            } else if (keyEvent->key() == Qt::Key_Down) {
                if (view->currentIndex().row() == 0) {
                    wid->ui->stashgroupTableView->selectRow(0);
                    wid->ui->stashgroupTableView->setFocus();
                    return true;
                }
            } else if (keyEvent->key() == Qt::Key_Tab) {
                view->selectionModel()->clearSelection();
                wid->ui->button_add->setFocus();
                return true;
            }
        }
        return QObject::eventFilter(obj, event);
    }
};


class StashGroupsNavigationFilter : public QObject
{
    Q_OBJECT
public:
    explicit StashGroupsNavigationFilter(QObject *parent = nullptr) : QObject(parent) {}

    bool eventFilter(QObject *obj, QEvent *event) {
        if (event->type() == QEvent::KeyPress) {
            StashgroupWidget *wid = qobject_cast<StashgroupWidget*>(parent());
            QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
            QTableView *view = qobject_cast<QTableView*>(obj);

            if (keyEvent->key() == Qt::Key_Up) {
                if (view->currentIndex().row() == 0) {
                    wid->ui->defaultTableView->selectRow(0);
                    wid->ui->defaultTableView->setFocus();
                    return true;
                }
            } else if (keyEvent->key() == Qt::Key_Tab) {
                view->selectionModel()->clearSelection();
                wid->ui->button_add->setFocus();
                return true;
            }
        }
        return QObject::eventFilter(obj, event);
    }
};


// This model shows the titlebar and the non-movable All Group =======================
class StashgroupHeadingModel : public QAbstractListModel
{
    Q_OBJECT

public:
    explicit StashgroupHeadingModel(QObject *parent = nullptr) : QAbstractListModel(parent) {}
    virtual int rowCount(const QModelIndex &parent) const { Q_UNUSED(parent) return 1; }

    virtual QVariant headerData(int /*section*/, Qt::Orientation /*orientation*/, int role) const override
    {
        if (role == Qt::DisplayRole) {
            return tr("Stash Groups");
        }
        return QVariant();
    }

    virtual QVariant data(const QModelIndex &/*index*/, int role) const
    {
        if (role == Qt::DisplayRole) {
            return tr("All");
        }
        return QVariant();
    }

    virtual Qt::ItemFlags flags(const QModelIndex &index) const override
    {
        Q_UNUSED(index)
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemNeverHasChildren;
    }
};


// This model shows the default stash group =======================================================
// Only allows drop from stashes onto the default group
class DefaultListModel : public QAbstractListModel
{
    Q_OBJECT

public:
    explicit DefaultListModel(QObject *parent = nullptr) : QAbstractListModel(parent) {}
    virtual int rowCount(const QModelIndex &parent) const { Q_UNUSED(parent) return 1; }

    virtual Qt::ItemFlags flags(const QModelIndex &index) const override
    {
        Qt::ItemFlags flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemNeverHasChildren;

        if (index.isValid() && draggedItemIsStash) {
            return flags | Qt::ItemIsDropEnabled;
        }
        return flags;
    }

    virtual QVariant data(const QModelIndex &/*index*/, int role) const
    {
        if (role == Qt::DisplayRole) {
            return tr("Default");
        }
        return QVariant();
    }

    virtual bool canDropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) const override
    {
        Q_UNUSED(action) Q_UNUSED(row) Q_UNUSED(column) Q_UNUSED(parent)

        if (data->formats().contains(QLatin1String("application/vnd.moritzmolch.encfsmanager.stash"))) {
            draggedItemIsStash = true;
            return true;
        } else {
            draggedItemIsStash = false;
            return false;
        }
    }

private:
    mutable bool draggedItemIsStash;

    // QAbstractItemModel interface
public:
    virtual bool dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) override
    {
        Q_UNUSED(action) Q_UNUSED(row) Q_UNUSED(column)

        if (data->formats().contains(QLatin1String("application/vnd.moritzmolch.encfsmanager.stash"))) {
            if (!parent.isValid()) {
                return false;
            }

            if ((parent.row() >= 0) && parent.row() < (Settings::instance().stashGroups().size())) {
                const StashMimeData *stashMimeData = qobject_cast<const StashMimeData*>(data);
                stashMimeData->stash()->setStashGroup(QLatin1String(""));
            }
        }
        return QAbstractItemModel::dropMimeData(data, action, row, column, parent);
    }

    virtual Qt::DropActions supportedDropActions() const override
    {
        return Qt::MoveAction;
    }
};


// This model shows the movable stashgroups =======================================================
class StashgroupModel : public QAbstractListModel
{
    Q_OBJECT

public:
    explicit StashgroupModel(QObject *parent = nullptr) : QAbstractListModel(parent) {
        connect(&Settings::instance(), &Settings::stashGroupsChanged, [this]() {
            beginResetModel();
            endResetModel();
        });
    }
    virtual int rowCount(const QModelIndex &parent) const { Q_UNUSED(parent) return Settings::instance().stashGroups().size(); }

    virtual Qt::DropActions supportedDropActions() const override { return Qt::MoveAction | Qt::CopyAction; }
    virtual Qt::DropActions supportedDragActions() const override { return Qt::MoveAction | Qt::CopyAction; }

    virtual QVariant data(const QModelIndex &index, int role) const
    {
        if (role == Qt::DisplayRole) {
            return Settings::instance().stashGroups().at(index.row());
        }
        return QVariant();
    }



    virtual Qt::ItemFlags flags(const QModelIndex &index) const override
    {
        if (index.isValid()) {
            lastValidRow = index.row();
        }
        if (lastValidRow < 0) {
            qDebug() << lastValidRow;
            return Qt::ItemIsEnabled;
        }

        Qt::ItemFlags flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemNeverHasChildren | Qt::ItemIsEditable | Qt::ItemIsDragEnabled;;

        if (fromStash) {
            return flags  | Qt::ItemIsDropEnabled;
        } else {
            if (!index.isValid()) {
                return flags | Qt::ItemIsDropEnabled;
            }
        }

        return flags;
    }

private:
    mutable int lastValidRow;
    mutable bool fromStash;

    // QAbstractItemModel interface
public:
    virtual bool canDropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) const override
    {
        Q_UNUSED(action) Q_UNUSED(row) Q_UNUSED(column) Q_UNUSED(parent)

        if (data->formats().contains(QLatin1String("application/vnd.moritzmolch.encfsmanager.stash"))) {
            fromStash = true;
        } else {
            fromStash = false;
        }

        if (fromStash || data->formats().contains(QLatin1String("application/x-qabstractitemmodeldatalist"))) {
            return true;
        }

        return false;
    }

    virtual bool dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) override
    {
        Q_UNUSED(action) Q_UNUSED(column)

        if (data->formats().contains(QLatin1String("application/vnd.moritzmolch.encfsmanager.stash"))) {
            if (!parent.isValid()) {
                return false;
            }

            if ((parent.row() >= 0) && parent.row() < (Settings::instance().stashGroups().size())) {
                const StashMimeData *stashMimeData = qobject_cast<const StashMimeData*>(data);
                stashMimeData->stash()->setStashGroup(Settings::instance().stashGroups().at(parent.row()));
            }


        } else if (data->formats().contains(QLatin1String("application/x-qabstractitemmodeldatalist"))) {
            if (row < 0) {
                return false;
            }

            int draggedRow = qobject_cast<StashgroupWidget*>(this->parent())->selectedStashGroupRow();

            if (row > draggedRow) {
                row--;
            }

            QStringList list = Settings::instance().stashGroups();
            list.move(draggedRow, row);

            Settings::instance().setStashGroups(list);
        }

        return QAbstractItemModel::dropMimeData(data, action, row, column, parent);
    }
friend class StashgroupWidget;
    // QAbstractItemModel interface
public:
    virtual bool setData(const QModelIndex &index, const QVariant &value, int role) override
    {
        Q_UNUSED(role)

        if (value.isNull() || value.toString().isEmpty()) {
            return false;
        }

        foreach (auto stash, StashManager::instance().stashes()) {
            if ((stash->stashGroup() == qobject_cast<StashgroupWidget*>(parent())->selectedStashGroup()) &&
                (!stash->stashGroup().isEmpty())) {
                stash->setStashGroup(value.toString());
            }
        }

        QStringList newList = Settings::instance().stashGroups();
        newList.replace(index.row(), value.toString());
        qobject_cast<StashgroupWidget*>(parent())->setSelectedStashGroup(value);
        Settings::instance().setStashGroups(newList);

        return true;
    }
};


StashgroupWidget::StashgroupWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::StashgroupWidget),
    m_selectedStashGroup(),
    m_selectedStashGroupRow(-1)
{
    ui->setupUi(this);
    ui->headerTableView->setModel(new StashgroupHeadingModel(this));
    ui->headerTableView->horizontalHeader()->setMinimumHeight(24);
    ui->headerTableView->selectRow(0);

    ui->defaultTableView->setModel(new DefaultListModel(this));
    ui->defaultTableView->setStyle(new MyProxyStyle(style()));

    ui->stashgroupTableView->setModel(new StashgroupModel(this));
    ui->stashgroupTableView->setStyle(new MyProxyStyle(style()));

    HeaderNavigationFilter *navFilter1 = new HeaderNavigationFilter(this);
    ui->headerTableView->installEventFilter(navFilter1);

    DefaultNavigationFilter *navFilter2 = new DefaultNavigationFilter(this);
    ui->defaultTableView->installEventFilter(navFilter2);

    StashGroupsNavigationFilter *navFilter3 = new StashGroupsNavigationFilter(this);
    ui->stashgroupTableView->installEventFilter(navFilter3);

    connect(ui->stashgroupTableView->model(), &StashgroupModel::modelReset, [this]() {
        if (m_selectedStashGroup.isNull()) {
            // header has focus
            ui->headerTableView->setCurrentIndex(ui->headerTableView->model()->index(0, 0));
            ui->defaultTableView->selectionModel()->clearSelection();
            ui->stashgroupTableView->selectionModel()->clearSelection();
        } else if (m_selectedStashGroup.toString().isEmpty()) {
            // default has focus
            ui->defaultTableView->setCurrentIndex(ui->headerTableView->model()->index(0, 0));
            ui->headerTableView->selectionModel()->clearSelection();
            ui->stashgroupTableView->selectionModel()->clearSelection();
        } else {
            int row = Settings::instance().stashGroups().indexOf(m_selectedStashGroup.toString());
            if (row >= 0) {
                ui->stashgroupTableView->setCurrentIndex(ui->stashgroupTableView->model()->index(row, 0));
            } else {
                ui->headerTableView->setCurrentIndex(ui->headerTableView->model()->index(0, 0));
                ui->defaultTableView->selectionModel()->clearSelection();
                ui->stashgroupTableView->selectionModel()->clearSelection();
            }

        }
    });


    connect(ui->headerTableView->selectionModel(), &QItemSelectionModel::selectionChanged,
            [this](const QItemSelection &selected, const QItemSelection &deselected) {
        Q_UNUSED(deselected)

        if (selected.size() > 0) {
            ui->stashgroupTableView->selectionModel()->clearSelection();
            ui->defaultTableView->selectionModel()->clearSelection();
            ui->headerTableView->setFocus();
        }
    });

    connect(ui->defaultTableView->selectionModel(), &QItemSelectionModel::selectionChanged,
            [this](const QItemSelection &selected, const QItemSelection &deselected) {
        Q_UNUSED(deselected)

        if (selected.size() > 0) {
            ui->headerTableView->selectionModel()->clearSelection();
            ui->stashgroupTableView->selectionModel()->clearSelection();
            ui->defaultTableView->setFocus();
        }
    });

    connect(ui->stashgroupTableView->selectionModel(), &QItemSelectionModel::selectionChanged,
            [this](const QItemSelection &selected, const QItemSelection &deselected) {
        Q_UNUSED(deselected)

        if (selected.size() > 0) {
            ui->headerTableView->selectionModel()->clearSelection();
            ui->defaultTableView->selectionModel()->clearSelection();
            ui->stashgroupTableView->setFocus();
        }
    });


    connect(&Settings::instance(), &Settings::managerShowStashGroupsChanged, [this](bool managerShowStashGroups) {
        if (!managerShowStashGroups) {
            ui->headerTableView->setCurrentIndex(ui->headerTableView->model()->index(0, 0));
        }
    });

    connect(ui->headerTableView->selectionModel(), &QItemSelectionModel::selectionChanged, [this](const QItemSelection &selected,
                                                                                                   const QItemSelection &/*deselected*/) {
        if (selected.indexes().isEmpty()) {
            return;
        }
        m_selectedStashGroupRow = -1;
        setSelectedStashGroup(QVariant());
    });

    connect(ui->defaultTableView->selectionModel(), &QItemSelectionModel::selectionChanged, [this](const QItemSelection &selected,
                                                                                                   const QItemSelection &/*deselected*/) {
        if (selected.indexes().isEmpty()) {
            return;
        }
        m_selectedStashGroupRow = -1;
        setSelectedStashGroup(QLatin1String(""));
    });

    connect(ui->stashgroupTableView->selectionModel(), &QItemSelectionModel::selectionChanged, [this](const QItemSelection &selected,
                                                                                                      const QItemSelection &/*deselected*/) {
        if (selected.indexes().isEmpty()) {
            return;
        }
        if ((selected.indexes().at(0).row() >= 0) && (selected.indexes().at(0).row() < Settings::instance().stashGroups().size())) {
            m_selectedStashGroupRow = selected.indexes().at(0).row();
            setSelectedStashGroup(Settings::instance().stashGroups().at(m_selectedStashGroupRow));
        }
    });
}

StashgroupWidget::~StashgroupWidget()
{
    delete ui;
}

QVariant StashgroupWidget::selectedStashGroup() const
{
    return m_selectedStashGroup;
}

int StashgroupWidget::selectedStashGroupRow() const
{
    return m_selectedStashGroupRow;
}

void StashgroupWidget::setSelectedStashGroup(const QVariant &selectedStashGroup)
{
    if (m_selectedStashGroup == selectedStashGroup) {
        return;
    }

    m_selectedStashGroup = selectedStashGroup;
    emit selectedStashGroupChanged(selectedStashGroup);
}

void StashgroupWidget::on_button_add_clicked()
{
    Settings::instance().setStashGroups(Settings::instance().stashGroups() + QStringList() << tr("New group"));
    QModelIndex index  = ui->stashgroupTableView->model()->index(Settings::instance().stashGroups().size()-1, 0);
    ui->stashgroupTableView->setCurrentIndex(index);
    ui->stashgroupTableView->edit(index);
}

void StashgroupWidget::on_button_remove_clicked()
{
    if (m_selectedStashGroupRow >=0 ) {
        QString stashGroup = Settings::instance().stashGroups().at(m_selectedStashGroupRow);

        foreach (auto stash, StashManager::instance().stashes()) {
            if (stash->stashGroup() == stashGroup) {
                stash->setStashGroup(QLatin1String(""));
            }
        }

        QStringList newList = Settings::instance().stashGroups();
        newList.removeAt(m_selectedStashGroupRow);
        Settings::instance().setStashGroups(newList);
    }
}

#include "stashgroupwidget.moc"
